/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.pc;

import java.util.ArrayList;

import javassist.CtClass;
import javassist.ClassPool;
import javassist.bytecode.Descriptor;
import javassist.compiler.CompileError;
import javassist.compiler.MemberResolver;
import javassist.gluonj.WeaveException;
import javassist.gluonj.weave.AdviceWeaver;

/*
 * Parse a given string.  The acceptable syntax
 *
 * pattern    : [<class name> ['+'] '#'] <member-or-new> ['(' <parameters> ')']
 *            |  <class name> ['+']
 *
 * member-or-new : <member name> | "new"
 *
 * parameters : <empty>
 *            | (<type name> ',')* (<type name> | '..')
 *
 * A wildcard '*' is available.
 */
public class Parser {
    static final char NOT = '!';
    static final char SUBCLASS = '+';
    static final char WILDCARD = '*';
    static final String ANY_PARAM = "..";
    static final String NEW = "new";
    static final String CONSTRUCTOR = AdviceWeaver.CONSTRUCTOR;

    private MemberResolver resolver;
    private String glueName;

    public Parser(String name, ClassPool cp) {
        resolver = new MemberResolver(cp);
        glueName = name;
    }

    public String getGlueName() { return glueName; }

    public PcPattern parseClass(String text) throws WeaveException {
        return parse0(text, true);
    }

    public PcPattern parse(String text) throws WeaveException {
        return parse0(text, false);
    }

    private PcPattern parse0(String text, boolean onlyClass)
        throws WeaveException
    {
        boolean notPat;
        String className, memberName, descriptor;
        int index;
        int textLen;
        if (text == null || (textLen = text.length()) <= 0)
            throw new WeaveException("empty pattern");

        if (text.charAt(0) == NOT) {
            notPat = true;
            index = 1;
        }
        else {
            notPat = false;
            index = 0;
        }

        int i = text.indexOf('#');
        if (i <= index)
            className = null;
        else {
            className = text.substring(index, i);
            index = i + 1;
        }

        if (index >= textLen) {
            memberName = null;
            descriptor = null;
        }
        else {
            i = text.indexOf('(', index);
            if (i < 0) {
                memberName = text.substring(index);
                descriptor = null;            
            }
            else {
                memberName = text.substring(index, i);
                index = i + 1;
                String[] paramList = parseParams(text, index);
                StringBuffer sbuf = new StringBuffer();
                makeDescriptor(sbuf, text, paramList);
                descriptor = sbuf.toString();
            }
        }

        if (className == null && onlyClass) {
            className = memberName;
            memberName = null;
        }

        boolean subclass = false;
        if (className != null) {
            int len = className.length();
            if (className.charAt(len - 1) == SUBCLASS) {
                if (len > 1) {
                    className = className.substring(0, len - 1);
                    subclass = true;
                }
                else
                    className = null;
            }
        }

        if (NEW.equals(memberName))
            memberName = CONSTRUCTOR;

        return new PcPattern(resolver, notPat, makeRegex(className),
                             subclass, makeRegex(memberName), descriptor);
    }

    private String makeRegex(String pat) {
        if (pat == null)
            return null;

        StringBuffer sbuf = new StringBuffer();
        int len = pat.length();
        for (int i = 0; i < len; i++) {
            char c = pat.charAt(i);
            if (c == '.') {
                sbuf.append('\\');
                sbuf.append('.');
            }
            else if (c == WILDCARD)
                sbuf.append(".*");
            else
                sbuf.append(c);
        }

        return sbuf.toString();
    }

    private void makeDescriptor(StringBuffer sbuf, String origText, String[] paramList)
        throws WeaveException
    {
        sbuf.append('(');
        for (int k = 0; k < paramList.length; k++) {
            String type = paramList[k];
            if (type.equals(ANY_PARAM))
                if (k + 1 == paramList.length)
                    return;
                else
                    throw new WeaveException("bad use of ..:" + origText);

            if (includesWildCards(type))
                throw new WeaveException("cannot use a wild card in a parameter list");
            else
                type = resolveClassName(type);

            toDescriptor(sbuf, type);
        }

        sbuf.append(')');
    }

    private static boolean includesWildCards(String text) {
        return text.indexOf(WILDCARD) >= 0;
    }

    private String resolveClassName(String className)
        throws WeaveException
    {
        try {
            CtClass cc = resolver.lookupClass(className, false);
            return cc.getName();
        }
        catch (CompileError e) {
            throw new WeaveException(e.getMessage());
        }
    }

    private static String[] parseParams(String text, int index) throws WeaveException {
        ArrayList list = new ArrayList();
        int len = text.length();
        StringBuffer sbuf = new StringBuffer();
        int c;
        while ((c = getNext(text, index++, len)) != ')') {
            if (c == ',') {
                if (sbuf.length() < 1)
                    throw new WeaveException("bad syntax: " + text);

                list.add(sbuf.toString());
                sbuf = new StringBuffer();
            }
            else if (c != ' ' && c != '\t')
                sbuf.append((char)c);
        }

        if (sbuf.length() > 0)
            list.add(sbuf.toString());

        return (String[])list.toArray(new String[list.size()]);
    }

    private static int getNext(String text, int i, int len) throws WeaveException {
        if (i < len)
            return text.charAt(i);
        else
            throw new WeaveException("bad syntax:" + text);
    }

    private static void toDescriptor(StringBuffer sbuf, String type) {
        int dim = 0;
        while (type.endsWith("[]")) {
            dim++;
            type = type.substring(0, type.length() - 2);
        }

        while (dim-- > 0)
            sbuf.append('[');

        sbuf.append(Descriptor.of(type));
    }
}
