/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.LoaderClassPath;
import javassist.NotFoundException;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.lang.instrument.ClassFileTransformer;
import java.lang.instrument.IllegalClassFormatException;
import java.lang.instrument.Instrumentation;
import java.security.ProtectionDomain;

import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;

public class HotSwapper {
    public static void premain(String args, Instrumentation inst)
        throws WeaveException
    {
        if (args == null || args.length() < 1)
            throw new WeaveException("no glue specified");

        inst.addTransformer(new Transformer(args));
    }

    public static class Transformer implements ClassFileTransformer {
        private String glueName;
        private ClassPool pool;
        private Weaver weaver;

        public Transformer(String glue) {
            glueName = glue;
            pool = null;
            weaver = null;
        }

        private void setupWeaver(ClassLoader cl) throws WeaveException {
            if (weaver != null)
                return;

            ClassPool cp = new ClassPool(true);
            if (cl != null)
                cp.insertClassPath(new LoaderClassPath(cl));

            pool = cp;
            Gluon g = Gluon.make(glueName, cp);
            weaver = new Weaver(g) {
                protected void save(CtClass clazz, boolean modify)
                    throws CannotCompileException, IOException {}
            };
        }

        public byte[] transform(ClassLoader loader, String className, Class classBeingRedefined,
                                ProtectionDomain domain, byte[] classfile)
            throws IllegalClassFormatException
        {
            if (isNonTransformable(className))
                return null;

            try {
                setupWeaver(loader);
                CtClass clazz = makeClass(className, classfile);
                if (weaver.transform(clazz.getName()))
                    return clazz.toBytecode();
                else
                    return classfile;
            }
            catch (Throwable t) {
                showError(className, t);    // this always throws an exception.
                return null;
            }
        }

        private CtClass makeClass(String className, byte[] classfile)
            throws IOException, WeaveException
        {
            int counter = 0;
            do {
                try {
                    return pool.makeClass(new ByteArrayInputStream(classfile));
                }
                catch (RuntimeException re) {}   // the class is already transformed.
                try {
                    pool.get(className.replace('/', '.')).detach();
                }
                catch (NotFoundException e) {
                    counter = 1;
                }
            } while (counter++ == 0);
            throw new WeaveException("cannot reload (broken class file?): "
                                     + className.replace('/', '.'));
        }

        private void showError(String className, Throwable e)
            throws RuntimeException
        {
            String msg = "cannot transform a class: " + className.replace('/', '.');
            System.err.println("GluonJ Error: " + msg);
            System.err.println("  by " + e);
            throw new RuntimeException(msg, e);
        }

        public boolean isNonTransformable(String jvmName) {
            return jvmName.startsWith("java/")
                || jvmName.startsWith("javax/")
                || jvmName.startsWith("com/sun/")
                || jvmName.startsWith("sun/")
                || jvmName.startsWith("sunw/")
                || jvmName.startsWith("javassist/");
        }
    }
}
