/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;

import javassist.ClassPool;
import javassist.CtClass;
import javassist.CannotCompileException;
import javassist.NotFoundException;
import javassist.Loader;
import javassist.gluonj.Glue;
import javassist.gluonj.Gluon;
import javassist.gluonj.Refine;
import javassist.gluonj.WeaveException;

public class OfflineWeaver extends Weaver {
    private String directory;
    private ArrayList notSavedClasses;

    /**
     * Constructs an offline weaver.
     * @throws WeaveException 
     */
    public OfflineWeaver(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        super(makeGlue(glueName, classpath));
        directory = ".";
        notSavedClasses = new ArrayList();
    }

    private static Gluon makeGlue(String glueName, String[] classpath)
        throws NotFoundException, WeaveException
    {
        ClassPool cp = new ClassPool(true);
        ClassPool cp2 = new ClassPool(true);
        if (classpath != null)
            for (int i = 0; i < classpath.length; i++) {
                String path = dirName(classpath[i]);
                cp.appendClassPath(path);
                cp2.appendClassPath(path);
            }

        Loader cloader = new Loader(Gluon.class.getClassLoader(), cp);
        cloader.delegateLoadingOf("javassist.");
        return Gluon.make(glueName, cp2, cloader);
    }

    private static String dirName(String dir) {
        String path = dir.replace(File.separatorChar, '/');
        int len = path.length();
        if (path.charAt(len - 1) == '/')
            path = path.substring(0, len - 1);

        return path;
    }

    /**
     * Sets the output directory where class files are written.
     */
    public void setDirectory(String dir) {
        directory = dirName(dir);
    }

    protected void save(CtClass clazz, boolean modify)
        throws CannotCompileException, IOException
    {
        if (modify)
            clazz.writeFile(directory);
        else
            notSavedClasses.add(clazz);
    }

    private void saveAll()
        throws CannotCompileException, IOException
    {
        if (directory.equals("."))
            return;

        Iterator it = notSavedClasses.iterator();
        while (it.hasNext()) {
            CtClass cc = (CtClass)it.next();
            cc.writeFile(directory);
        }
    }

    /**
     * Transforms class files.
     */
    public void transformFile(String fileName) throws WeaveException {
        ClassPool cp = getClassPool();
        readFile(cp, null, fileName);
    }

    /**
     * Transforms class files.
     */
    public void transform(File baseDir, String[] fileNames) throws WeaveException {
        int len = fileNames.length;
        notSavedClasses.ensureCapacity(len);
        ClassPool cp = getClassPool();
        for (int i = 0; i < len; i++) {
            String fname = fileNames[i];
            if (fname.endsWith(".class"))
                readFile(cp, baseDir, fname);
        }

        Exception ex;
        try {
            saveAll();
            return;
        }
        catch (IOException e) { ex = e; }
        catch (CannotCompileException e) { ex = e; }
        throw new WeaveException(ex);
    }

    private void readFile(ClassPool cp, File baseDir, String name)
        throws WeaveException
    {
        try {
            InputStream is
                = new BufferedInputStream(new FileInputStream(new File(baseDir, name)));
            CtClass clazz = cp.makeClass(is);
            is.close();
            if (notGlue(clazz))
                transform(clazz.getName());
        }
        catch (IOException e) {
            throw new WeaveException(e);
        }
    }

    private boolean notGlue(CtClass clazz) {
        boolean result = true;
        Object[] anno;
        try {
            anno = clazz.getAnnotations();
        }
        catch (ClassNotFoundException e) {
            return true;
        }

        for (int j = 0; j < anno.length; j++)
            if (anno[j] instanceof Glue || anno[j] instanceof Refine)
                result = false;

        return result;
    }

    public static void main(String[] args) throws Exception {
        if (args.length < 1) {
            showUsage();
            return;
        }

        OfflineWeaver w = new OfflineWeaver(args[0], null);
        w.setDirectory("./out/");
        for (int i = 1; i < args.length; i++)
           w.transformFile(args[i]);
    }

    private static void showUsage() {
        System.out.println("Usage: java OfflineWeaver "
                + "<glue class name> <java class file> ...");
    }
}
