/*
 * GluonJ, An Aspect-Oriented Programming System for Java.
 * Copyright (C) 2006 Shigeru Chiba. All Rights Reserved.
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License.  Alternatively, the contents of this file may be used under
 * the terms of the GNU Lesser General Public License Version 2.1 or later.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 */
package javassist.gluonj.weave;

import javassist.CannotCompileException;
import javassist.ClassPool;
import javassist.CtClass;
import javassist.NotFoundException;
import javassist.gluonj.Gluon;
import javassist.gluonj.WeaveException;

import java.io.IOException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;

/**
 * A glue processor.
 */
public abstract class Weaver {
    public static interface Factory {
        RefineWeaver refineWeaver();
        AdviceWeaver adviceWeaver();
    }

    /**
     * This value can be changed for extending a weaver.
     */
    public static Factory factory = new Factory() {
        public RefineWeaver refineWeaver() {
            return new RefineWeaver();
        }

        public AdviceWeaver adviceWeaver() {
            return new AdviceWeaver();
        }
    };

    private Gluon glue;
    protected ClassPool pool;
    private RefineWeaver refineWeaver;
    private AdviceWeaver adviceWeaver;
    protected HashMap refineTargets;

    /**
     * Constructs a weaver.
     */
    public Weaver(Gluon g) {
        glue = g;
        pool = g.getPool();
        refineWeaver = factory.refineWeaver();
        adviceWeaver = factory.adviceWeaver();
        refineTargets = new HashMap();
        loadRefines();
    }

    private void loadRefines() {
        Iterator iterator = glue.getRefiners();
        while (iterator.hasNext()) {
            String[] pair = (String[])iterator.next();
            addRefine(pair[0], pair[1]);
        }
    }

    // Don't use this method; it is not private since it is called from test code.
    void addRefine(String target, String refine) {
        LinkedList entry = (LinkedList)refineTargets.get(target);
        if (entry == null)
            entry = new LinkedList();

        entry.addLast(refine);
        refineTargets.put(target, entry);
    }

    protected ClassPool getClassPool() {
        return glue.getPool();
    }

    /**
     * Transforms a given class.
     *
     * @param targetName        fully-qualified class name.
     */
    protected boolean transform(String targetName) throws WeaveException {
        return transform(targetName, (LinkedList)refineTargets.get(targetName));
    }

    /**
     * Applies given refine classes to a target class
     * and then weaves advices.
     *
     * @param refines       a list of refine classes obtained from
     *                      refineTargets.
     * @return              true if the target class was modified. 
     */
    protected boolean transform(String targetName, LinkedList refines)
        throws WeaveException
    {
        try {
            CtClass target = pool.get(targetName);
            boolean modify = refineClass(refineWeaver, target, refines);
            boolean modify2 = weaveAdvices(target);
            boolean result = modify || modify2; 
            save(target, result);
            return result;
        }
        catch (IOException ioe) {
            throw new WeaveException(ioe);
        }
        catch (CannotCompileException cce) {
            throw new WeaveException(cce);
        }
        catch (NotFoundException nfe) {
            throw new WeaveException(nfe);
        }
    }

    /**
     * Applies refine classes to a target class.
     *
     * @param refines   a list of refine classes.
     */
    protected boolean refineClass(RefineWeaver rw, CtClass target,
                                  LinkedList refines)
        throws WeaveException, NotFoundException, CannotCompileException
    {
        if (refines == null)
            return false;

        boolean modified = false;
        Iterator it2 = refines.iterator();
        while (it2.hasNext()) {
            String refine = (String)it2.next();
            target = rw.weave(target, pool.get(refine));
            modified = true;
        }

        return modified;
    }

    /**
     * Weaves advices into a target class.
     */
    protected boolean weaveAdvices(CtClass target)
        throws WeaveException, CannotCompileException
    {
        return adviceWeaver.weave(target, glue);
    }

    /**
     * Finalizes the transformation of a given class.
     * This method must be implemented in a subclass.
     *
     * @param modify    the class was modified.
     */
    protected abstract void save(CtClass clazz, boolean modify)
        throws CannotCompileException, IOException;
}
